from datetime import datetime, timedelta

import logging
import time
import traceback
from dataclasses import dataclass
from multiprocessing import Process
from typing import Callable

from core.domain.optimization.OptimizationState import OptimizationState
from core.domain.worker.Worker import Worker
from core.domain.worker.WorkerJob import WorkerJob
from core.domain.worker.WorkerLogLevel import WorkerLogLevel
from core.domain.worker.WorkerState import WorkerState
from core.domain.worker.WorkerType import WorkerType
from core.repos.OptimizationRepo import OptimizationRepo
from core.repos.WorkerJobRepo import WorkerJobRepo
from core.repos.WorkerLogRepo import WorkerLogRepo
from core.repos.WorkerRepo import WorkerRepo
from core.services.SystemService import SystemService
from core.types.Id import Id
from core.usecases.initialization.Register_worker import Register_worker

log = logging.getLogger(__name__)


@dataclass
class Run_optimization_worker:
	register_worker: Register_worker
	workerLogRepo: WorkerLogRepo
	optimizationRepo: OptimizationRepo
	workerJobRepo: WorkerJobRepo
	systemService: SystemService
	init_run_optimization_job: Callable[[WorkerJob], None]
	init_log_worker_status: Callable[[Id[Worker]], None]
	workerRepo: WorkerRepo

	def getActiveOptimizationsWarmingUp(self, stateChangedDuration: timedelta):
		activeOptimizations = self.optimizationRepo.getWithState(state=OptimizationState.ACTIVE)
		return list(filter(lambda opt: (datetime.now() - opt.stateChangedAt) < stateChangedDuration, activeOptimizations))

	def now(self):
		log.info("Register worker")
		result = self.register_worker.now(workerType=WorkerType.OPTIMIZER)
		workerId = result.worker.id

		log.info("Create new daemon process for the worker logging")
		job = Process(target=self.init_log_worker_status, kwargs={'id': workerId}, name=result.worker.name)
		job.daemon = False
		job.start()

		log.info("Waiting for worker events...")
		while True:
			try:
				log.debug("===[END OPTIMIZATION WORKER]=============================")
				log.debug("")
				log.debug("")
				log.debug("")
				time.sleep(5)
				log.debug("===[RUN OPTIMIZATION WORKER]=============================")

				worker = self.workerRepo.get(id=workerId)
				if worker.state == WorkerState.DEPRECATED:
					log.info(f"Worker {worker.id} is deprecated")
					continue

				maxRamOptimizationReserved = 7000
				availableCpu = self.systemService.getCpuAvailable()
				maxMbRam = self.systemService.getMaxRamMbAvailable()
				availableMbRam = self.systemService.getRamMbAvailable()
				takenMbRam = maxMbRam - availableMbRam
				activeOptimizationsWarmingUp = self.getActiveOptimizationsWarmingUp(stateChangedDuration=timedelta(hours=1))
				virtualTakenMbRam = takenMbRam + (maxRamOptimizationReserved * len(activeOptimizationsWarmingUp))
				virtualFreeMbRam = maxMbRam - virtualTakenMbRam

				log.info(", ".join([
					f"Worker {worker.id}: {worker.state.name}",
					f"CPU={availableCpu}",
					f"WARMUP={len(activeOptimizationsWarmingUp)}",
					f"VIR_FREE_RAM={round(virtualFreeMbRam / 1e3, 1)}Gb",
					f"FREE_RAM={round(availableMbRam / 1e3, 1)}Gb",
					f"MAX_RAM={round(maxMbRam / 1e3, 1)}Gb",
				]))

				if availableCpu == 0:
					log.warn("System is missing available cpus")
					continue

				if virtualFreeMbRam < maxRamOptimizationReserved:
					log.warn("System is missing available ram")
					continue

				log.debug("Get first optimization and update it into active state")
				activeOptimization = self.optimizationRepo.updateFirst(fromState=OptimizationState.CREATED, toState=OptimizationState.ACTIVE)

				if activeOptimization is None:
					log.debug(f"Active optimization not found")
					continue

				log.info(f"Active optimization found")
				log.info("Create new worker job")
				workerJob = WorkerJob(
					workerId=workerId,
					optimizationId=activeOptimization.id,
					name=activeOptimization.title,
					state=OptimizationState.CREATED
				)
				self.workerJobRepo.post(obj=workerJob)

				log.info("Create new daemon process for the optimization")
				job = Process(target=self.init_run_optimization_job, kwargs={'workerJob': workerJob},
				              name=f"{result.worker.name} {activeOptimization.type.value} '{workerJob.name}'")
				job.daemon = False
				job.start()

				log.info(f"Job process: PID={job.pid}, NAME={job.name}, DEAMON={job.daemon}")

			except BaseException as err:
				log.error(err)
				log.error(traceback.format_exc())
				self.workerLogRepo.post(context=__name__, workerId=workerId, data=f"Exception: {err}", level=WorkerLogLevel.ERROR)

		job.terminate()
